/******************************************************************************
* FILE: switchtests.c
* DESCRIPTION:
*   Conducts bandwidth and latency tests for MPI point-to-point message
*   passing.
* SOURCE: Blaise Barney  5/99
* LAST REVISED: 
****************************************************************************/

#include "mpi.h"
#include <stdio.h>
#include <sys/time.h>
#include <time.h>

#define L_REPS   1000000
#define B_REPS   1000
#define TBYTES   3000000

struct      timeval tv1, tv2;    
int         i, n, rc, thistime, besttime, worsttime, totaltime, tag, dest, 
            source, numtasks, taskid, len;
float       bandwidth, bestbandwidth, worstbandwidth, totalbandwidth;
char        inmsg[1], outmsg[1], a[TBYTES], hostname[MPI_MAX_PROCESSOR_NAME];
MPI_Status  status;

int main(argc,argv)
int argc;
char *argv[]; {

MPI_Init(&argc,&argv);
MPI_Comm_size(MPI_COMM_WORLD,&numtasks);
MPI_Comm_rank(MPI_COMM_WORLD,&taskid);

/*** Some sanity checks ***/
if (numtasks < 2) {
  printf("ERROR: Number of tasks= %d\n", numtasks);
  printf("Need 2 tasks to run this test.  Quitting \n");
  MPI_Abort(MPI_COMM_WORLD,rc);
  }
else if (numtasks > 2) {
  printf("INFO: Number of tasks= %d\n", numtasks);
  printf("Only need 2 tasks to run this test. Ignoring extra tasks. \n");
  }

/*** Some initializations ***/
outmsg[0] = 'x';
for (i=0; i<TBYTES; i++) 
  a[i] = 'x';
tag = 1;
MPI_Get_processor_name(hostname, &len);
printf("Using machine: %s\n", hostname);

/****************************** task 0 work *********************************/
if (taskid == 0) {

  /*** Find out how much overhead is incurred from timing routines ***/
  besttime = 999999;
  worsttime = 0;
  totaltime = 0;
  for (i=0; i<L_REPS; i++) {
    gettimeofday(&tv1, (struct timeval*)0);
    gettimeofday(&tv2, (struct timeval*)0);
    thistime = (tv2.tv_sec - tv1.tv_sec) * 1000000 + tv2.tv_usec - tv1.tv_usec;
    totaltime = totaltime + thistime;
    if ((thistime < besttime) && (thistime > 0)) besttime = thistime;
    if (thistime > worsttime) worsttime = thistime;
    }
  printf("\nOverhead times (best/avg/worst) =  %d / %d / %d uSec\n\n", 
          besttime, totaltime/L_REPS, worsttime);

  /*** Do rountrip latency test ***/
  printf("Starting round trip latency test: message size=1, reps= %d\n",L_REPS);
  dest = 1;
  source = 1;
  besttime=999999;
  worsttime = 0;
  totaltime = 0;
  for (n=0; n<L_REPS; n++) {
    gettimeofday(&tv1, (struct timeval*)0);
    MPI_Send(outmsg, 1, MPI_BYTE, dest, tag, MPI_COMM_WORLD);
    MPI_Recv(inmsg, 1, MPI_BYTE, source, tag, MPI_COMM_WORLD, &status);
    gettimeofday(&tv2, (struct timeval*)0);
    thistime = (tv2.tv_sec - tv1.tv_sec) * 1000000 + tv2.tv_usec - tv1.tv_usec;
    totaltime = totaltime + thistime;
    if ((thistime < besttime) && (thistime > 0)) besttime = thistime;
    if (thistime > worsttime) worsttime = thistime;
    } 
  printf("Roundtrip times (best/avg/worst) =  %d / %d / %d uSec\n\n", 
          besttime, totaltime/L_REPS, worsttime);

  /*** Do bandwidth test ***/
  printf("Starting bandwidth test: message size= %d, reps= %d\n",TBYTES,B_REPS);
  bestbandwidth = 0.0;
  worstbandwidth = 999999999999.0;
  totalbandwidth = 0.0; 
  for (n=0; n<B_REPS; n++) {
    gettimeofday(&tv1, (struct timeval*)0);
    MPI_Send(a, TBYTES, MPI_BYTE, dest, tag, MPI_COMM_WORLD);
    MPI_Recv(a, TBYTES, MPI_BYTE, source, tag, MPI_COMM_WORLD, &status);
    gettimeofday(&tv2, (struct timeval*)0);
    thistime = (tv2.tv_sec - tv1.tv_sec) * 1000000 + tv2.tv_usec - tv1.tv_usec;
    bandwidth = ((float)TBYTES / ((float)thistime/1000000.0)) * 2.0 ;
    totalbandwidth = totalbandwidth + bandwidth;
    if (bandwidth > bestbandwidth) bestbandwidth = bandwidth;
    if (bandwidth < worstbandwidth) worstbandwidth = bandwidth;
    }
  printf("Bandwidth results (best/avg/worst) = %4.2f / %4.2f / %4.2f MB/sec \
         \n\n", bestbandwidth/1000000.0, 
          (totalbandwidth/1000000.0)/(float)B_REPS, worstbandwidth/1000000.0);
  }

/****************************** task 1 work *********************************/
if (taskid == 1) {
  dest = 0;
  source = 0;
  for (n=0; n<L_REPS; n++) {
    MPI_Recv(inmsg, 1, MPI_BYTE, source, tag, MPI_COMM_WORLD, &status);
    MPI_Send(outmsg, 1, MPI_BYTE, dest, tag, MPI_COMM_WORLD);
    }

  for (n=0; n<B_REPS; n++) {
    MPI_Recv(a, TBYTES, MPI_BYTE, source, tag, MPI_COMM_WORLD, &status);
    MPI_Send(a, TBYTES, MPI_BYTE, dest, tag, MPI_COMM_WORLD);
    }

  }

MPI_Finalize();

}
