C -----------------------------------------------------------------------
C  MPI Concurrent Wave Equation  - Fortran Example
C  FILE: mpi_wave.f
C  OTHER FILES: mpi_wave.h, draw_wave.c
C  DESCRIPTION:
C    This program implements the concurrent wave equation described 
C    in Chapter 5 of Fox et al., 1988, Solving Problems on Concurrent
C    Processors, vol 1.  
C
C    A vibrating string is decomposed into points.  Each processor is 
C    responsible for updating the amplitude of a number of points over
C    time.
C	
C    At each iteration, each processor exchanges boundary points with
C    nearest neighbors.  This version uses low level sends and receives
C    to exchange boundary points.
C
C LAST REVISED: 12/20/2001 Blaise Barney
C  ------------------------------------------------------------------------ 
C  ------------------------------------------------------------------------ 
C  Explanation of constants and variables used in common blocks and
C  include files
C    MASTER           = task ID of master
C    TPOINTS          = total points along wave
C    OUT1, OUT2       = message types
C    LtoR, RtoL       = message types
C    taskid           = task ID
C    numtasks         = number of tasks
C    nsteps           = number of time steps
C    npoints          = number of points handled by this task
C    first            = index of first point handled by this task
C    values(0:1001)   = values at time t
C    oldval(0:1001)   = values at time (t-dt)
C    newval(0:1001)   = values at time (t+dt)
C  ------------------------------------------------------------------------ 

      program wave_send
      
      include 'mpi_wave.h'
      include 'mpif.h'

C     Routine for creating the X graph of the wave
      external draw_wave
      integer left, right, ierr

C     Determine number of tasks and taskid
      call MPI_INIT(ierr)
      call MPI_COMM_RANK(MPI_COMM_WORLD, taskid, ierr)
      call MPI_COMM_SIZE(MPI_COMM_WORLD, numtasks, ierr)
      if (numtasks .lt. 2) then
         print *, 'ERROR: Number of MPI tasks set to' ,numtasks
         print *, 'Need at least 2 tasks. Quitting...'
         call MPI_ABORT(MPI_COMM_WORLD, ierr)
         stop
      end if

C     Determine left and right neighbors
      if (taskid .eq. numtasks-1) then
         right = 0
      else 
         right = taskid + 1
      end if

      if (taskid .eq. 0) then
         left = numtasks - 1
      else 
         left = taskid - 1
      end if

C     Get program parameters and initialize wave values
      if (taskid .eq. MASTER) then
         print *,'MPI_WAVE starting. Using',numtasks,'tasks.'
         print *,'Using',TPOINTS,'points on the vibrating string.'
         call init_master
      else
         call init_workers
      end if
      call init_line

C     Update values along the wave for nstep time steps
      call update(left, right)

C     Master collects results from workers and prints
      if (taskid .eq. MASTER) then
         call out_master
      else 
         call out_workers
      end if
      call MPI_FINALIZE(ierr)
      end

C  ------------------------------------------------------------------------
C     Master obtains timestep input value from user and broadcasts it
C  ------------------------------------------------------------------------ 
      subroutine init_master

      include 'mpi_wave.h'
      include 'mpif.h'
      integer ierr

      nsteps = 0
      do while ((nsteps .lt. 1) .or. (nsteps .gt. MAXSTEPS))
         write (*,*) 'Enter number of time steps (1-',MAXSTEPS,'):'
         read (*,*) nsteps
         if ((nsteps .lt. 1) .or. (nsteps .gt. MAXSTEPS))
     &        write (*,*) 'enter value between 1 and ', MAXSTEPS
      end do

      call MPI_BCAST(nsteps, 1, MPI_INTEGER, MASTER, MPI_COMM_WORLD,
     &               ierr)
      end

C  -------------------------------------------------------------------------
C     Workers receive timestep input value from master
C  -------------------------------------------------------------------------
      subroutine init_workers

      include 'mpi_wave.h'
      include 'mpif.h'
      integer ierr

      call MPI_BCAST(nsteps, 1, MPI_INTEGER, MASTER, MPI_COMM_WORLD,
     &               ierr)

      end

C  ------------------------------------------------------------------------
C     Initialize points on line
C  -----------------------------------------------------------------------
      subroutine init_line

      include 'mpi_wave.h'

      integer nmin, nleft, npts, i, j, k
      real*8 x, fac

C     Calculate initial values based on sine curve
      nmin = TPOINTS/numtasks
      nleft = mod(TPOINTS, numtasks)
      fac = 2.0 * PI

      k = 0
      do i = 0, numtasks-1
         if (i .lt. nleft) then
            npts = nmin + 1
         else
            npts = nmin
         endif
         if (taskid .eq. i) then
            first = k + 1
            npoints = npts
	    write (*,15) taskid, first, npts
 15         format ('task=',I5, ': first = ', I5, ' npoints = ', I5)
            do j = 1, npts
               x = float(k)/float(TPOINTS - 1)
               values(j) = sin (fac * x)
               k = k + 1
            end do
         else 
            k = k + npts
         end if
      end do
      do i = 1, npoints
         oldval(i) = values(i)
      end do
      end

C  -------------------------------------------------------------------------
C     Update all values along line a specified number of times 
C  -------------------------------------------------------------------------
      subroutine update(left, right)

      include 'mpi_wave.h'
      include 'mpif.h'

      integer left, right, ierr, status(MPI_STATUS_SIZE), i, j
      real*8 dtime, c, dx, tau, sqtau

      dtime = 0.3
      c = 1.0
      dx = 1.0
      tau = (c * dtime / dx)
      sqtau = tau * tau

C     Update values for each point along string
      do i = 1, nsteps

C        Exchange data with "left-hand" neighbor
         if (first .ne. 1) then
            call MPI_SEND(values(1), 1, MPI_DOUBLE_PRECISION, left,
     &                    RtoL, MPI_COMM_WORLD, ierr)
            call MPI_RECV(values(0), 1, MPI_DOUBLE_PRECISION, left,
     &                     LtoR, MPI_COMM_WORLD, status, ierr)
         end if
C        Exchange data with "right-hand" neighbor
	 if (first+npoints-1 .ne. TPOINTS) then
            call MPI_SEND(values(npoints), 1, MPI_DOUBLE_PRECISION,
     &               right, LtoR, MPI_COMM_WORLD, ierr)
            call MPI_RECV(values(npoints+1), 1, MPI_DOUBLE_PRECISION,
     &                 right, RtoL, MPI_COMM_WORLD, status, ierr)
         end if

C        Update points along line
         do j = 1, npoints
C           Global endpoints
            if ((first+j-1 .eq. 1).or.(first+j-1 .eq. TPOINTS))then
               newval(j) = 0.0
	        else
C              Use wave equation to update points
               newval(j) = (2.0 * values(j)) - oldval(j)
     &         + (sqtau * (values(j-1) - (2.0 * values(j)) 
     &         + values(j+1)))
            end if
         end do

         do j = 1, npoints
            oldval(j) = values(j)
	    values(j) = newval(j)
         end do

      end do

      end

C  ------------------------------------------------------------------------
C     Receive results from workers and print
C  ------------------------------------------------------------------------ 
      subroutine out_master

      include 'mpi_wave.h'
      include 'mpif.h'

      integer i, j, start, npts, buffer(2), 
     &        ierr, status(MPI_STATUS_SIZE)
      real*8 results(TPOINTS) 
 
C     Store worker's results in results array
      do i = 1, numtasks - 1
C        Receive first point, number of points and results
         call MPI_RECV(buffer, 2, MPI_INTEGER, i, OUT1,
     &                 MPI_COMM_WORLD, status, ierr)
         start = buffer(1)
         npts = buffer(2)
         call MPI_RECV(results(start), npts, MPI_DOUBLE_PRECISION,
     &                 i, OUT2, MPI_COMM_WORLD, status, ierr )
      end do

C     Store master's results in results array
      do i = first, first+npoints-1
         results(i) = values(i)
      end do

      j = 0
      print *,'********************************************************'
      print *,'Final amplitude values for all points after',nsteps,
     &        ' steps:'
      do i = 1, TPOINTS, 10
         write(*,FMT='(10F6.2)') (results(j),j=i,i+9)
      end do
      print *, ' '
      print *, 'Click the EXIT button or use CTRL-C to quit'

C     Display results with draw_wave routine
      call draw_wave(%REF(results))

      end

C  -------------------------------------------------------------------------
C     Send the updated values to the master
C  -------------------------------------------------------------------------
      subroutine out_workers

      include 'mpi_wave.h'
      include 'mpif.h'

      integer ierr, buffer(2)

C     Send first point and number of points handled to master
      buffer(1) = first
      buffer(2) = npoints
      call MPI_SEND(buffer, 2, MPI_INTEGER, MASTER, OUT1, 
     &              MPI_COMM_WORLD, ierr)

C     Send results to master
      call MPI_SEND(values(1), npoints, MPI_DOUBLE_PRECISION, MASTER,
     &              OUT2, MPI_COMM_WORLD, ierr)

      end
